import numpy as np
from scipy import stats as st
import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf

from orthog_model import OrthogonalSymmetryCayley as osc
from normal_model import NormalModel as nm
from experiments import eval_and_plot_orthog
from utils import cosine_similarity

from kmm import kernel_mean_matching

def gaussian_boundaries(mean, cov, df, n_samples):
    """
    Generates samples from an m-dimensional Gaussian distribution and splits
    them based on a random selection vector.
    """
    m = mean.shape[0]
    gaussian_samples = np.random.multivariate_normal(mean, cov, n_samples)

    a = np.random.rand(m, 1) - 0.5
    b = np.random.uniform(-1, 1)

    gaussian_side = np.dot(gaussian_samples, a) > b

    gaussian_selected = gaussian_samples[gaussian_side.flatten()]
    gaussian_not_selected = gaussian_samples[~gaussian_side.flatten()]

    return a, b, gaussian_selected, gaussian_not_selected

def t_boundaries(mean, cov, df, n_samples):
    """
    Generates samples from an m-dimensional t-distribution and splits
    them based on a random selection vector.
    """
    m = mean.shape[0]
    t_samples = st.multivariate_t.rvs(mean, cov, df, size=n_samples)

    a = np.random.rand(m, 1) - 0.5
    b = np.random.uniform(-1, 1)

    t_side = np.dot(t_samples, a) > b

    t_selected = t_samples[t_side.flatten()]
    t_not_selected = t_samples[~t_side.flatten()]

    return a, b, t_selected, t_not_selected

def generate_random_mean(m):
    """
    Generates a random mean vector of dimension m.

    :param m: Dimensionality of the mean vector.
    :return: A random mean vector.
    """
    mean = np.random.rand(m) - 0.5
    return mean
def generate_random_covariance(m):
    """
    Generates a random positive-definite covariance matrix of size m x m.
    """
    A = np.random.rand(m, m) - 0.5
    A = (A + A.T) / 2
    cov = A + m * np.eye(m)
    return cov
def generate_random_degrees_of_freedom():
    """
    Generates a random integer for degrees of freedom.
    """
    df = np.random.randint(low=3, high=30)
    return df

def cosine_similarity_experiment(m=3, num_exp=1000):
    normal_gauss_sims = []
    normal_t_sims = []

    orthog_gauss_sims = []
    orthog_t_sims = []
    for i, data_gen in enumerate([gaussian_boundaries, t_boundaries]):
        for _ in range(num_exp):
            print(i)
            mean = generate_random_mean(m)
            cov = generate_random_covariance(m)
            df = generate_random_degrees_of_freedom()
            a, b, data_sel, data_unsel = data_gen(mean, cov, df, 2000)
            data = np.concatenate((data_sel, data_unsel), axis=0)
            X = tf.convert_to_tensor(
                data_sel,
                dtype=tf.float32)
            # Normal
            model = nm(M=m)
            model.train(X)
            w = model.params[2].numpy()
            if i == 0:
                normal_gauss_sims.append(cosine_similarity(a, w))
            else:
                normal_t_sims.append(cosine_similarity(a, w))

            # Orthogonal
            model = osc(X)
            model.train()
            U = model.U
            scores = tf.reduce_sum(model.gaussian_kernel(
                X, tf.matmul(X, U, transpose_b=True)),
                                    axis=1).numpy()
            w, _ = eval_and_plot_orthog(data, data_sel, U, scores, 'normal_test', plot=False)
            if i == 0:
                orthog_gauss_sims.append(cosine_similarity(a, w))
            else:
                orthog_t_sims.append(cosine_similarity(a, w))

            # KMM
            u = st.ortho_group.rvs(m)
            U = np.dot(u, u.T)
            a2 = U @ a
            data_side = np.dot(data, a2) > b
            data_sel_2 = data[data_side.flatten()]
            kernel_mean_matching(data_sel, data_sel_2)

    return normal_gauss_sims, normal_t_sims, orthog_gauss_sims, orthog_t_sims

def plot_similarities(sims_1, sims_2, sims_3, sims_4):
    plt.figure(figsize=(10, 6))
    for i, sims in enumerate([sims_1, sims_2, sims_3, sims_4]):
        sorted_sims = sorted(sims, reverse=True)
        c = ['red', 'blue', 'green', 'yellow'][i]
        plt.plot(sorted_sims, marker='o', linestyle='-', c=c, alpha=0.5)
        plt.title('Cosine Similarities in Descending Order')
        plt.xlabel('Experiment Number')
        plt.ylabel('Cosine Similarity')
        plt.grid(True)
    plt.show()

if __name__ == '__main__':
    plot_similarities(*cosine_similarity_experiment())
