import numpy as np
import pandas as pd
import tensorflow as tf
from scipy import stats as st
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression
import exodata

from orthog_model import OrthogonalSymmetryCayley as osc
from normal_model import NormalModel as nm
from kmm import kernel_mean_matching


def pengu_experiment():
    pengu = sns.load_dataset('penguins')
    plt.figure(figsize=(12,5))
    for q in [0.66]:
        for p in pengu.species.unique():
            pengu_species_ = pengu[pengu.species==p]
            pengu_sel = pengu_species_[pengu_species_.body_mass_g <
                                       pengu_species_.body_mass_g.quantile(q)]
            pengu_species = pengu_species_[['flipper_length_mm', 'bill_length_mm']]
            pengu_sel = pengu_sel[['flipper_length_mm', 'bill_length_mm']]
            pengu_sel_tensor = tf.convert_to_tensor(
                pengu_sel.select_dtypes(include=[float]).values,
                dtype=tf.float32)
            # Normal model
            model = nm(M=2)
            model.train(pengu_sel_tensor)
            a = model.params[2].numpy()
            b = model.params[3].numpy()
            eval_and_plot_normal(pengu_species, pengu_sel, a, b, f'Penguins: {p}')

            # Orthogonal model
            model = osc(pengu_sel_tensor)
            model.train()
            U = model.U
            scores = tf.reduce_sum(model.gaussian_kernel(
                pengu_sel_tensor, tf.matmul(pengu_sel_tensor, U, transpose_b=True)),
                                   axis=1).numpy()
            eval_and_plot_orthog(pengu_species, pengu_sel, U, scores, f'Penguins: {p}')

            # KMM
            u = st.ortho_group.rvs(2)
            U = np.dot(u, u.T)
            pengu_sel_2 = pengu_species_[pengu_species_.body_mass_g <
                                         pengu_species_.body_mass_g.quantile(q+0.05)]
            kernel_mean_matching(pengu_sel, pengu_sel_2)


def exoplanet_experiment():
    exocat = exodata.load_db_from_url()
    Ls = [star.estimateAbsoluteMagnitude() for star in exocat.stars]
    Ds = []
    for star in exocat.stars:
        try:
            Ds.append(star.d.item())
        except AttributeError:
            Ds.append(star.d)
            df = pd.DataFrame([Ds, Ls]).T
            df = df.dropna(thresh=2)
            df = df[np.log10(df.iloc[:, 0]) > 1]
            df = df[np.log10(df.iloc[:, 0]) < 3.5]
            df.iloc[:, 0] = np.log10(df.iloc[:, 0])
            X = tf.convert_to_tensor(
                df.select_dtypes(include=[float]).values,
                dtype=tf.float32)

    # Normal model
    model = nm(M=2)
    model.train(X)
    a = model.params[2].numpy()
    b = model.params[3].numpy()
    eval_and_plot_normal(X, X, a, b, 'Exoplanets')

    # Orthogonal model
    model = osc(X)
    model.train()
    U = model.U
    scores = tf.reduce_sum(model.gaussian_kernel(
        X, tf.matmul(X, U, transpose_b=True)),
                           axis=1).numpy()
    eval_and_plot_orthog(X, X, U, scores, 'Exoplanets')

def eval_and_plot_orthog(X, X_sel, scores, name, plot=True):
    try:
        X = X.values
    except:
        pass
    labels = np.where(scores > np.median(scores), 1, 0)
    X_train, _, y_train, _ = train_test_split(X_sel, labels,
                                              test_size=0.1, random_state=42)
    lr = LogisticRegression(max_iter=1000)
    lr.fit(X_train, y_train)
    y_pred = lr.predict(X_sel)

    w = lr.coef_[0]
    b = lr.intercept_[0]

    if plot:
        # Points predicted as mapping into selected away area
        X_sel_pred_1 = X_sel[y_pred == 1]
        # Points predicted as not mapping into selected away area
        X_sel_pred_0 = X_sel[y_pred == 0]

        plt.figure(figsize=(10, 6))

        plt.scatter(X[:, 0], X[:, 1], color='green',
                    label='Selected away points', alpha=0.5)
        plt.scatter(X_sel_pred_1[:, 0], X_sel_pred_1[:, 1],
                    color='blue', label='Selected away region', alpha=0.5)
        plt.scatter(X_sel_pred_0[:, 0], X_sel_pred_0[:, 1],
                    color='red', label='Not selected away region', alpha=0.5)


        x_values = np.linspace(X[:, 0].min(), X[:, 0].max(), 100)
        y_values = -(w[0]/w[1]) * x_values - b/w[1]
        plt.plot(x_values, y_values, color="black", linestyle="--",
                 label='Decision Boundary')

        plt.title(f'Orthogonal result for {name}')
        plt.ylim(X[:, 1].min(), X[:, 1].max())
        plt.legend()
        plt.savefig(f'figs/{name}_orthog.png', dpi=360)
    return w, b

def eval_and_plot_normal(X, X_sel, a, b, name, plot=True):
    a = np.array(a).reshape(-1, 1)

    X_dot_a = np.dot(X.values, a)

    selected_indices = X_dot_a.flatten() > 0  
    X_sel_pred_1 = X_sel

    x_values = np.linspace(X.iloc[:, 0].min(), X.iloc[:, 0].max(), 100)
    y_values = -(a[0]/a[1]) * x_values - b/a[1]

    if plot:
        plt.figure(figsize=(10, 6))
        plt.scatter(X.iloc[:, 0], X.iloc[:, 1], color='green',
                    label='All points', alpha=0.5)
        plt.scatter(X_sel_pred_1.iloc[:, 0], X_sel_pred_1.iloc[:, 1],
                    color='blue', label='Selected away region', alpha=0.5)
        plt.plot(x_values, y_values, color="black", linestyle="--",
                 label='Decision Boundary')

        plt.ylim(X.iloc[:, 1].min(), X.iloc[:, 1].max())

        plt.title(f'Normal result for {name}')
        plt.legend()
        plt.savefig(f'figs/{name}_normal.png', dpi=360)


if __name__ == '__main__':
    exoplanet_experiment()
    pengu_experiment()
