import os.path
import glob

import matplotlib
import matplotlib.pyplot as plt
import pickle
import itertools as it
import seaborn as sns
import networkx as nx

import pandas as pd
import numpy as np
import scipy as sp
import scipy.stats as st
import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow import keras
from keras import backend as K
tfd = tfp.distributions

from dcd.admg_discovery import Discovery
from kmm import kernel_mean_matching


class NormalModel(object):
    def __init__(self, params=None, M=2):
        self.params = params
        self.trainable = self.params
        self.optimizer = keras.optimizers.Adam()
        self.M = M
        if self.params is None:
            self.param_init()
        else:
            self.params = params

    def param_init(self):
        mu_init = np.random.normal(0, 1, size=(self.M,))
        sigma_init = np.diag(np.random.lognormal(0, 1, self.M))
        a_init = np.random.normal(0, 1, size=(self.M,))
        b_init = np.random.uniform(-1, 1)
        zeta_init = np.random.uniform(0.1, 1.)

        mu_ = tf.Variable(mu_init, dtype=tf.dtypes.float32, name='mu')
        sigma_ = tf.Variable(sigma_init, dtype=tf.dtypes.float32, name='cov')
        a_ = tf.Variable(a_init.reshape(self.M, 1), dtype=tf.dtypes.float32, name='a')
        b_ = tf.Variable(b_init, dtype=tf.dtypes.float32, name='b')
        zeta_ = tf.Variable(zeta_init, dtype=tf.dtypes.float32, name='zeta')
        self.params = [mu_, sigma_, a_, b_, zeta_]
        self.trainable = self.params

    def loss(self, X):
        N = X.shape[0]
        mean, cov, a, b, zeta = self.params
        cov_sqrt = tf.linalg.cholesky(cov)
        cov_sqrt = tf.cast(cov_sqrt, tf.dtypes.float32)
        distr_data = tfd.MultivariateNormalTriL(loc=mean, scale_tril=cov_sqrt)
        distr_cutoff = tfd.Normal(loc=0, scale=1)

        self.log_pdf = -tf.math.reduce_sum(distr_data.log_prob(X))
        self.log_prob_misclassified = -tf.math.reduce_sum(distr_cutoff.log_cdf((tf.matmul(X, a) + b)/zeta))
        self.log_prob_included = N*tf.math.reduce_sum(distr_cutoff.log_cdf((
            tf.matmul(tf.reshape(mean, [1, self.M]), a) + b)/tf.math.sqrt(tf.norm(tf.matmul(cov_sqrt, a))**2 + zeta**2)))
        log_loss = self.log_pdf + self.log_prob_misclassified + self.log_prob_included
        log_loss = log_loss + tf.norm(a)**2 + tf.norm(b)**2 - tfd.InverseGamma(0.1, 0.1).log_prob(zeta)
        self.log_loss = log_loss
        return log_loss

    def training_step(self, batch, training_vars, k):
        for _ in range(k):
            with tf.GradientTape() as tape:
                L = self.loss(batch)
                grad = tape.gradient(L, training_vars)
                self.optimizer.apply_gradients(zip(grad, training_vars))

    def train(self, X, epochs=100, batch_size=64):
        N = X.shape[0]
        [mu_, sigma_, a_, b_, zeta_] = self.params

        training_vars = [mu_, sigma_, a_, b_, zeta_]

        for epoch in range(epochs):
            for _ in range(N // batch_size):
                indices = tf.random.shuffle(tf.range(N))[:batch_size]
                batch = tf.gather(X, indices)
                self.training_step(batch, training_vars, k=1)

        self.params = [mu_, sigma_, a_, b_, zeta_]
